/* This file is part of Mailfromd.
   Copyright (C) 2006, 2007, 2008, 2009, 2010 Sergey Poznyakoff

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 3, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program.  If not, see <http://www.gnu.org/licenses/>. */

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif
#include <stdlib.h>
#include <stdio.h>
#include "mailfromd.h"

struct module_level {
	char *name;
	size_t level;
	int isset;
};

static struct module_level *modtab;
static size_t modmax;
static size_t modcnt;
#define DEFAULT_MODMAX 16;

size_t
debug_register(char *name)
{
	if (modcnt == modmax) {
		if (modmax == 0)
			modmax = DEFAULT_MODMAX;
		modtab = x2nrealloc(modtab, &modmax, sizeof(modtab[0]));
	}
	modtab[modcnt].name = name;
	modtab[modcnt].level = 0;
	return modcnt++;
}

void
debug_init(char **names)
{
	debug_register("all");
	for (; *names; names++)
		debug_register(*names);
}

void
debug_freeze()
{
	if (modmax > modcnt) {
		modtab = xrealloc(modtab, modcnt * sizeof(modtab[0]));
		modmax = modcnt;
	}
}

int
debug_level_p(int modn, int level)
{
	return modn < modcnt
		&& (modtab[modn].isset ? (modtab[modn].level >= level)
		    : modtab[0].level >= level);
}

static int
find_module(const char *name)
{
	size_t len, i;
	const char *q;
	const char *p;

	p = strrchr(name, '/');                     
	if (p)                                   
		p++;                             
	else                                     
		p = name;                           
	q = strrchr(p, '.');                     
	if (q)                                   
		len = q - name;                       
	else                                     
		len = strlen(p);

	for (i = 0; i < modcnt; i++) {
		if (strlen(modtab[i].name) == len
		    && memcmp(modtab[i].name, p, len) == 0)
			return i;
	}
	return -1;
}

void
debug_enable_module(const char *file, int level)
{
	int modn = find_module(file);
	if (modn < 0) {
		mu_error(_("unknown module: %s"), file);
		return;
	}
	modtab[modn].level = level;
	modtab[modn].isset = 1;
}

void
debug_disable_module(const char *file)
{
	int modn = find_module(file);
	if (modn < 0) {
		mu_error(_("unknown module: %s"), file);
		return;
	}
	modtab[modn].isset = 0;
}

int
debug_module_level(const char *modname, int *plev)
{
	int modn;

	if (modname) {
		modn = find_module(modname);
		if (modn < 0) 
			return 1;
	} else
		modn = 0;
	*plev = modtab[modn].level;
	return 0;
}

static void
parse_spec(char *spec)
{
	char *q;
	int level;

	for (;; spec++) {
		if (*spec == 0)
			return;
		if (!isspace(*spec))
			break;
	}

	if (isdigit(*spec)) {
		level = strtoul(spec, &q, 0);
		if (*q) 
			mu_error(_("%s: wrong debug spec near %s"), spec, q);
		modtab[0].level = level;
		if (modtab[0].level >= 100)
 			enable_prog_trace("all");
		return;
	}

	if (*spec == '!')
		debug_disable_module(spec + 1);
	else {
		char *p;

		p = strchr(spec, '=');
		if (p) {
			*p++ = 0;
			level = strtol(p, &q, 0);
			if (*q || level < 0) { 
				mu_error(_("%s: wrong debug spec near %s"),
					 spec, q);
				return;
			}
		} else
			level = 100;
		debug_enable_module(spec, level);
	}
}

void
debug_parse_spec(const char *spec)
{
	char *copy = xstrdup(spec);
	char *s;
	
	for (s = strtok(copy, ","); s; s = strtok(NULL, ","))
		parse_spec(s);
	free(copy);
}

int
debug_spec_string(const char *spec, int minlevel, int showunset, char **pbuf)
{
	char *mods;
	size_t len = 0, i;
	char nbuf[NUMERIC_BUFSIZE_BOUND];
	char *buf, *ptr;

	mods = malloc(modcnt);
	if (!mods) {
		errno = ENOMEM;
		return -1;
	}
		
	if (!spec || !*spec || strcmp(spec, "any") == 0) {
		for (i = 0; i < modcnt; i++)
			mods[i] = 1;
	} else {
		char *copy = xstrdup(spec);
		char *s;
	
		memset(mods, 0, modcnt);
		for (s = strtok(copy, ","); s; s = strtok(NULL, ",")) {
			int n;

			while (*s && isspace(*s))
				s++;
			n = find_module(s);
			if (n < 0)
				continue;
			mods[n] = 1;
		}
	}

	if (mods[0]) {
		snprintf(nbuf, sizeof nbuf, "%d", modtab[0].level);
		len += strlen(nbuf) + 1;
	}
	for (i = 1; i < modcnt; i++) {
		if (mods[i]) {
			if (modtab[i].isset) {
				if (modtab[i].level >= minlevel) {
					snprintf(nbuf, sizeof nbuf,
						 "%d", modtab[i].level);
					len += strlen(modtab[i].name) +
						1 +
						strlen(nbuf) + 1;
				}
			} else if (showunset)
				len += strlen(modtab[i].name) + 2;
		}
	}		

	buf = malloc(len + 1);
	if (!buf) {
		free(mods);
		errno = ENOMEM;
		return -1;
	}
	
	ptr = buf;
	if (mods[0] && modtab[0].level >= minlevel) {
		sprintf(ptr, "%d", modtab[0].level);
		ptr += strlen(ptr);
	}
	for (i = 1; i < modcnt; i++) {
		if (mods[i]) {
			if (modtab[i].isset) {
				if (ptr > buf)
					*ptr++ = ',';
				if (modtab[i].level >= minlevel) {
					sprintf(ptr, "%s=%d",
						modtab[i].name,
						modtab[i].level);
					ptr += strlen(ptr);
				}
			} else if (showunset) {
				if (ptr > buf)
					*ptr++ = ',';
				sprintf(ptr, "!%s", modtab[i].name);
				ptr += strlen(ptr);
			}
		}
	}
	*ptr = 0;
	free(mods);
	
	*pbuf = buf;
	return 0;
}
